@@ -1,16 +1,318 @@
 import dataclasses
 import gc
+import math
 import os
+import pickle
 import random
+import struct
 from collections import defaultdict
 from dataclasses import asdict, dataclass
-from typing import Any, Iterator, List, Optional, Union
+from datetime import datetime
+from typing import Any, ByteString, Callable, Iterator, List, Optional, Tuple, Union
 
 import lmdb
+import msgpack
 import numpy
+import psutil
+import ujson
+from lz4 import frame
+from pyroaring import BitMap
 from tqdm import tqdm
 
-from freaddb import config, io_worker, utils
+ENCODING = "utf-8"
+SIZE_1GB = 1_073_741_824  # 1GB
+LMDB_MAX_KEY = 511
+LMDB_MAP_SIZE = SIZE_1GB
+LMDB_BUFF_LIMIT = SIZE_1GB
+
+
+class ToBytes:
+    OBJ = 0
+    INT_NUMPY = 1
+    INT_BITMAP = 2
+    BYTES = 3
+    PICKLE = 4
+
+
+class DBUpdateType:
+    SET = 0
+    COUNTER = 1
+
+
+def create_dir(file_dir: str):
+    folder_dir = os.path.dirname(file_dir)
+    if not os.path.exists(folder_dir):
+        os.makedirs(folder_dir)
+
+
+def save_json_file(file_name: str, json_obj: dict, indent: int = 2):
+    with open(file_name, "w") as f:
+        ujson.dump(json_obj, f, indent=indent)
+
+
+def read_json_file(file_name: str):
+    with open(file_name, "r") as f:
+        return ujson.load(f)
+
+
+def is_byte_obj(obj: Any) -> bool:
+    if isinstance(obj, bytes) or isinstance(obj, bytearray):
+        return True
+    return False
+
+
+def deserialize_key(
+    key: Any,
+    integerkey=False,
+    is_64bit=False,
+    combinekey: bool = False,
+    combinelen: int = 2,
+    **kawgs,
+) -> Union[int, str]:
+
+    # String key
+    if not integerkey:
+        if isinstance(key, memoryview):
+            key = key.tobytes()
+        return key.decode(ENCODING)
+
+    # Integer key
+    format_template = "Q" if is_64bit else "I"
+    # Tuple integer key
+    if combinekey:
+        format_template = format_template * combinelen
+        return struct.unpack(format_template, key)
+
+    # Single integer key
+    return struct.unpack(format_template, key)[0]
+
+
+def deserialize_value(
+    value: ByteString,
+    bytes_value: ToBytes = ToBytes.OBJ,
+    compress_value: bool = False,
+    **kawgs,
+) -> Any:
+    if bytes_value == ToBytes.INT_NUMPY:
+        value = numpy.frombuffer(value, dtype=numpy.uint32)
+
+    elif bytes_value == ToBytes.INT_BITMAP:
+        if not isinstance(value, bytes):
+            value = bytes(value)
+        value = BitMap.deserialize(value)
+
+    elif bytes_value == ToBytes.BYTES:
+        if isinstance(value, memoryview):
+            value = value.tobytes()
+
+    else:  # mode == "msgpack"
+        if compress_value:
+            try:
+                value = frame.decompress(value)
+            except RuntimeError:
+                pass
+        if bytes_value == ToBytes.PICKLE:
+            value = pickle.loads(value)
+        else:
+            value = msgpack.unpackb(value, strict_map_key=False)
+    return value
+
+
+def deserialize(
+    key: ByteString,
+    value: ByteString,
+    integerkey: bool = False,
+    combinekey: bool = False,
+    combinelen: int = 2,
+    is_64bit: bool = False,
+    bytes_value: ToBytes = ToBytes.OBJ,
+    compress_value: bool = False,
+    **kawgs,
+) -> Tuple[Any, Any]:
+    key = deserialize_key(
+        key=key,
+        integerkey=integerkey,
+        is_64bit=is_64bit,
+        combinekey=combinekey,
+        combinelen=combinelen,
+    )
+    value = deserialize_value(
+        value=value, bytes_value=bytes_value, compress_value=compress_value
+    )
+    res_obj = (key, value)
+    return res_obj
+
+
+def serialize_key(
+    key: Any,
+    integerkey: bool = False,
+    combinekey: bool = False,
+    is_64bit: bool = False,
+    combinelen: int = 2,
+    **kawgs,
+) -> ByteString:
+    if not integerkey:
+        if not isinstance(key, str):
+            key = str(key)
+        return key.encode(ENCODING)[:LMDB_MAX_KEY]
+
+    # Integer key
+    format_template = "Q" if is_64bit else "I"
+    # Tuple integer key
+    if combinekey:
+        format_template = format_template * combinelen
+        return struct.pack(format_template, *key)
+
+    if (
+        not isinstance(key, int)
+        and not isinstance(key, list)
+        and not isinstance(key, tuple)
+    ):
+        raise TypeError
+
+    return struct.pack(format_template, key)
+
+
+def serialize_value(
+    value: Any,
+    bytes_value: ToBytes = ToBytes.OBJ,
+    compress_value: bool = False,
+    sort_values: bool = True,
+    **kawgs,
+) -> ByteString:
+    def set_default(obj):
+        if isinstance(obj, set):
+            return sorted(list(obj))
+        raise TypeError
+
+    if bytes_value == ToBytes.INT_NUMPY:
+        if sort_values:
+            value = sorted(list(value))
+        if not isinstance(value, numpy.ndarray):
+            value = numpy.array(value, dtype=numpy.uint32)
+        value = value.tobytes()
+
+    elif bytes_value == ToBytes.INT_BITMAP:
+        value = BitMap(value).serialize()
+
+    else:  # mode == "msgpack"
+        if bytes_value == ToBytes.PICKLE:
+            value = pickle.dumps(value)
+        else:
+            if not isinstance(value, bytes) and not isinstance(value, bytearray):
+                value = msgpack.packb(value, default=set_default)
+        if compress_value:
+            value = frame.compress(value)
+
+    return value
+
+
+def serialize(
+    key: Any,
+    value: Any,
+    integerkey: bool = False,
+    combinekey: bool = False,
+    combinelen: int = 2,
+    is_64bit: bool = False,
+    bytes_value: ToBytes = ToBytes.OBJ,
+    compress_value: bool = False,
+    **kawgs,
+) -> Tuple[ByteString, ByteString]:
+    key = serialize_key(
+        key=key,
+        integerkey=integerkey,
+        combinekey=combinekey,
+        combinelen=combinelen,
+        is_64bit=is_64bit,
+    )
+    value = serialize_value(
+        value=value, bytes_value=bytes_value, compress_value=compress_value
+    )
+    res_obj = (key, value)
+    return res_obj
+
+
+def preprocess_data_before_dump(
+    data: List[Any],
+    integerkey: bool = False,
+    combinekey: bool = False,
+    combinelen: int = 2,
+    is_64bit: bool = False,
+    bytes_value: ToBytes = ToBytes.OBJ,
+    compress_value: bool = False,
+    sort_key: bool = True,
+    **kwargs,
+) -> List[Any]:
+    if isinstance(data, dict):
+        data = list(data.items())
+
+    if sort_key and integerkey:
+        data.sort(key=lambda x: x[0])
+
+    first_key, first_value = data[0]
+    to_bytes_key = not is_byte_obj(first_key)
+    to_bytes_value = not is_byte_obj(first_value)
+
+    for i in range(len(data)):
+        k, v = data[i]
+        if k is None:
+            continue
+        if to_bytes_key:
+            data[i][0] = serialize_key(
+                key=k,
+                integerkey=integerkey,
+                combinekey=combinekey,
+                combinelen=combinelen,
+                is_64bit=is_64bit,
+            )
+        if to_bytes_value:
+            data[i][1] = serialize_value(
+                value=v,
+                bytes_value=bytes_value,
+                compress_value=compress_value,
+            )
+
+    if sort_key and not integerkey:
+        data.sort(key=lambda x: x[0])
+
+    if not isinstance(data[0], tuple):
+        data = [(k, v) for k, v in data]
+    return data
+
+
+def get_file_size(num: int, suffix="B"):
+    num = abs(num)
+    if num == 0:
+        return "0B"
+    try:
+        magnitude = int(math.floor(math.log(num, 1024)))
+        val = num / math.pow(1024, magnitude)
+        if magnitude > 7:
+            return "{:3.1f}{}{}".format(val, "Yi", suffix)
+        return "{:3.1f}{}{}".format(
+            val, ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"][magnitude], suffix
+        )
+    except ValueError:
+        print(num)
+        return "0B"
+
+
+def profile(func: Callable):
+    def wrapper(*args, **kwargs):
+        process = psutil.Process(os.getpid())
+        mem_before = process.memory_info()
+        start = datetime.now()
+
+        result = func(*args, **kwargs)
+
+        end = datetime.now() - start
+        mem_after = process.memory_info()
+        rss = get_file_size(mem_after.rss - mem_before.rss)
+        vms = get_file_size(mem_after.vms - mem_before.vms)
+        print(f"{func.__name__}\tTime: {end}\tRSS: {rss}\tVMS: {vms}")
+        return result
+
+    return wrapper
 
 
 @dataclass
@@ -18,7 +320,7 @@ class DBSpec:
     name: str
     integerkey: bool = False
     is_64bit: bool = False
-    bytes_value: bool = config.ToBytes.OBJ
+    bytes_value: bool = ToBytes.OBJ
     compress_value: bool = False
     combinekey: bool = False
     combinelen: int = 2
@@ -31,15 +333,15 @@ class FReadDB:
         db_schema: Optional[List[DBSpec]] = None,
         map_size: int = None,
         readonly: bool = False,
-        buff_limit: int = config.LMDB_BUFF_LIMIT,
+        buff_limit: int = LMDB_BUFF_LIMIT,
         split_subdatabases=False,
     ):
 
         self.db_file = db_file
-        io_worker.create_dir(self.db_file)
+        create_dir(self.db_file)
 
         self.metadata_file = db_file + ".json"
-        io_worker.create_dir(self.metadata_file)
+        create_dir(self.metadata_file)
 
         if readonly and not os.path.exists(self.db_file) and not split_subdatabases:
             split_subdatabases = True
@@ -67,10 +369,10 @@ class FReadDB:
             "db_schema": [asdict(db_i) for db_i in db_schema],
             "buff_limit": buff_limit,
         }
-        io_worker.save_json_file(self.metadata_file, json_obj)
+        save_json_file(self.metadata_file, json_obj)
 
     def load_metadata_info(self):
-        json_obj = io_worker.read_json_file(self.metadata_file)
+        json_obj = read_json_file(self.metadata_file)
         db_schema = {obj["name"]: DBSpec(**obj) for obj in json_obj["db_schema"]}
         buff_limit = json_obj["buff_limit"]
         return db_schema, buff_limit
@@ -114,7 +416,7 @@ class FReadDB:
                     self.env[db_spec.name].set_mapsize(self.map_size)
 
             self.dbs[db_spec.name] = self.env[db_spec.name].open_db(
-                db_spec.name.encode(config.ENCODING), integerkey=db_spec.integerkey
+                db_spec.name.encode(ENCODING), integerkey=db_spec.integerkey
             )
 
         return True
@@ -130,7 +432,7 @@ class FReadDB:
 
         if not tmp:
             return "Unknown"
-        return utils.get_file_size(tmp)
+        return get_file_size(tmp)
 
     def get_db_size(self, db_name, pretty: bool = True):
         tmp = self.env[db_name].info().get("map_size")
@@ -138,7 +440,7 @@ class FReadDB:
             return tmp
         if not tmp:
             return "Unknown"
-        return utils.get_file_size(tmp)
+        return get_file_size(tmp)
 
     def get_number_items_from(self, db_name: str):
         with self.env[db_name].begin(db=self.dbs[db_name]) as txn:
@@ -150,6 +452,29 @@ class FReadDB:
             for env_i in self.env.values():
                 env_i.close()
 
+    def compress_subdatabase(self, db_name: str):
+        if not self.split_subdatabases:
+            raise ValueError(
+                "Error: Cannot compress subdatabase, please use compress function"
+            )
+        org_dir = self.db_file + f"_{db_name}"
+        new_dir = org_dir + ".copy"
+
+        size_org = os.stat(org_dir).st_size
+        self.env[db_name].copy(path=new_dir, compact=True)
+        try:
+            if os.path.exists(org_dir):
+                os.remove(org_dir)
+        except Exception as message:
+            print(message)
+        os.rename(new_dir, org_dir)
+        size_curr = os.stat(org_dir).st_size
+
+        print(
+            f"{db_name} : {100 - size_curr / size_org *100:.2f}% - {get_file_size(size_curr)}/{get_file_size(size_org)}"
+        )
+        return size_org, size_curr
+
     def compress(self) -> None:
         """
         Copy current env to new one (reduce file size)
@@ -169,31 +494,16 @@ class FReadDB:
             os.rename(new_dir, self.db_file)
             size_curr = os.stat(self.db_file).st_size
             print(
-                f"Compressed: {100 - size_curr / size_org *100:.2f}% - {utils.get_file_size(size_curr)}/{utils.get_file_size(size_org)}"
+                f"Compressed: {100 - size_curr / size_org *100:.2f}% - {get_file_size(size_curr)}/{get_file_size(size_org)}"
             )
         else:
             total_org, total_cur = 0, 0
             for db_name in self.db_schema.keys():
-                org_dir = self.db_file + f"_{db_name}"
-                new_dir = org_dir + ".copy"
-
-                size_org = os.stat(org_dir).st_size
+                size_org, size_curr = self.compress_subdatabase(db_name)
                 total_org += size_org
-                self.env[db_name].copy(path=new_dir, compact=True)
-                try:
-                    if os.path.exists(org_dir):
-                        os.remove(org_dir)
-                except Exception as message:
-                    print(message)
-                os.rename(new_dir, org_dir)
-                size_curr = os.stat(org_dir).st_size
-
                 total_cur += size_curr
-                print(
-                    f"{db_name} : {100 - size_curr / size_org *100:.2f}% - {utils.get_file_size(size_curr)}/{utils.get_file_size(size_org)}"
-                )
             print(
-                f"Compressed: {100 - total_cur / total_org *100:.2f}% - {utils.get_file_size(total_cur)}/{utils.get_file_size(total_org)}"
+                f"Compressed: {100 - total_cur / total_org *100:.2f}% - {get_file_size(total_cur)}/{get_file_size(total_org)}"
             )
 
     def get_random_key(self, db_name) -> Any:
@@ -202,10 +512,10 @@ class FReadDB:
             random_index = random.randint(0, self.get_number_items_from(db_name))
             cur = txn.cursor()
             cur.first()
-            key = utils.deserialize_key(cur.key(), **db_schema)
+            key = deserialize_key(cur.key(), **db_schema)
             for i, k in enumerate(cur.iternext(values=False)):
                 if i == random_index:
-                    key = utils.deserialize_key(k, **db_schema)
+                    key = deserialize_key(k, **db_schema)
                     break
         return key
 
@@ -217,13 +527,13 @@ class FReadDB:
             if to_i == -1:
                 to_i = self.get_number_items_from(db_name)
             cur = txn.cursor()
-            cur.set_range(utils.serialize_key(from_i, **db_schema))
+            cur.set_range(serialize_key(from_i, **db_schema))
             for item in cur.iternext(values=get_values):
                 if get_values:
                     key, value = item
                 else:
                     key = item
-                key = utils.deserialize_key(key, **db_schema)
+                key = deserialize_key(key, **db_schema)
                 if not isinstance(key, int):
                     raise ValueError(
                         f"This function used for integerkey databases. This is {type(key)} key database"
@@ -231,7 +541,7 @@ class FReadDB:
                 if key > to_i:
                     break
                 if get_values:
-                    value = utils.deserialize_value(value, **db_schema)
+                    value = deserialize_value(value, **db_schema)
                     yield key, value
                 else:
                     yield key
@@ -244,19 +554,19 @@ class FReadDB:
 
         with self.env[db_name].begin(db=self.dbs[db_name], write=False) as txn:
             cur = txn.cursor()
-            prefix = utils.serialize_key(prefix, **db_schema)
+            prefix = serialize_key(prefix, **db_schema)
             cur.set_range(prefix)
 
             while cur.key().startswith(prefix) is True:
                 try:
                     if cur.key() and not cur.key().startswith(prefix):
                         continue
-                    key = utils.deserialize_key(
+                    key = deserialize_key(
                         cur.key(),
                         **db_schema,
                     )
                     if get_values:
-                        value = utils.deserialize_value(cur.value(), **db_schema)
+                        value = deserialize_value(cur.value(), **db_schema)
                         yield key, value
                     else:
                         yield key
@@ -267,7 +577,7 @@ class FReadDB:
     def is_available(self, db_name: str, key_obj: str) -> bool:
         db_schema = dataclasses.asdict(self.db_schema[db_name])
         with self.env[db_name].begin(db=self.dbs[db_name]) as txn:
-            key_obj = utils.serialize_key(key_obj, **db_schema)
+            key_obj = serialize_key(key_obj, **db_schema)
             if key_obj:
                 try:
                     value_obj = txn.get(key_obj)
@@ -280,7 +590,7 @@ class FReadDB:
     def get_value_byte_size(self, db_name: str, key_obj: Any) -> Union[int, None]:
         db_schema = dataclasses.asdict(self.db_schema[db_name])
         with self.env[db_name].begin(db=self.dbs[db_name], buffers=True) as txn:
-            key_obj = utils.serialize_key(key_obj, **db_schema)
+            key_obj = serialize_key(key_obj, **db_schema)
             if key_obj:
                 try:
                     value_obj = txn.get(key_obj)
@@ -304,14 +614,14 @@ class FReadDB:
             ):
                 return responds
 
-            key_objs = [utils.serialize_key(k, **db_schema) for k in key_objs]
+            key_objs = [serialize_key(k, **db_schema) for k in key_objs]
             for k, v in txn.cursor(self.dbs[db_name]).getmulti(key_objs):
                 if not v:
                     continue
-                k = utils.deserialize_key(k, **db_schema)
+                k = deserialize_key(k, **db_schema)
                 if get_deserialize:
                     try:
-                        v = utils.deserialize_value(v, **db_schema)
+                        v = deserialize_value(v, **db_schema)
                     except Exception as message:
                         print(message)
                 responds[k] = v
@@ -321,7 +631,7 @@ class FReadDB:
     def get_value(self, db_name: str, key_obj: Any, get_deserialize: bool = True):
         db_schema = dataclasses.asdict(self.db_schema[db_name])
         with self.env[db_name].begin(db=self.dbs[db_name], buffers=True) as txn:
-            key_obj = utils.serialize_key(key_obj, **db_schema)
+            key_obj = serialize_key(key_obj, **db_schema)
             responds = None
             if not key_obj:
                 return responds
@@ -331,7 +641,7 @@ class FReadDB:
                     return responds
                 responds = value_obj
                 if get_deserialize:
-                    responds = utils.deserialize_value(value_obj, **db_schema)
+                    responds = deserialize_value(value_obj, **db_schema)
 
             except Exception as message:
                 print(message)
@@ -372,9 +682,9 @@ class FReadDB:
                     key = db_obj
                 try:
                     if deserialize_obj:
-                        key = utils.deserialize_key(key, **db_schema)
+                        key = deserialize_key(key, **db_schema)
                         if get_values:
-                            value = utils.deserialize_value(value, **db_schema)
+                            value = deserialize_value(value, **db_schema)
                     if get_values:
                         return_obj = (key, value)
                         yield return_obj
@@ -407,7 +717,7 @@ class FReadDB:
         ) as txn:
             for k in key:
                 try:
-                    status = txn.delete(utils.serialize_key(k, **db_schema))
+                    status = txn.delete(serialize_key(k, **db_schema))
                     if status:
                         deleted_items += 1
                 except Exception as message:
@@ -418,7 +728,7 @@ class FReadDB:
     def write(
         env, db, data, sort_key: bool = True, one_sample_write: bool = False, **kargs
     ):
-        data = utils.preprocess_data_before_dump(data, sort_key=sort_key, **kargs)
+        data = preprocess_data_before_dump(data, sort_key=sort_key, **kargs)
         added_items = 0
         try:
             with env.begin(db=db, write=True, buffers=True) as txn:
@@ -430,7 +740,7 @@ class FReadDB:
                         added_items += 1
         except lmdb.MapFullError:
             curr_limit = env.info()["map_size"]
-            new_limit = curr_limit + config.LMDB_BUFF_LIMIT
+            new_limit = curr_limit + LMDB_BUFF_LIMIT
             env.set_mapsize(new_limit)
             return FReadDB.write(env, db, data, sort_key=False)
         except lmdb.BadValsizeError:
@@ -459,14 +769,14 @@ class FReadDB:
         message: str = "DB Write",
         **kwargs,
     ) -> bool:
-        data = utils.preprocess_data_before_dump(
+        data = preprocess_data_before_dump(
             data,
             sort_key=sort_key,
             **kwargs,
         )
 
         def update_desc():
-            return f"{message} buffer: {buff_size / config.LMDB_BUFF_LIMIT * 100:.0f}%"
+            return f"{message} buffer: {buff_size / LMDB_BUFF_LIMIT * 100:.0f}%"
 
         p_bar = None
         buff_size = 0
@@ -480,7 +790,7 @@ class FReadDB:
                 p_bar.set_description(desc=update_desc())
             buff_size += len(k) + len(v)
 
-            if buff_size >= config.LMDB_BUFF_LIMIT:
+            if buff_size >= LMDB_BUFF_LIMIT:
                 c = FReadDB.write(env, db, data[i_pre:i], sort_key=False)
                 if c != len(data[i_pre:i]):
                     print(
@@ -502,11 +812,11 @@ class FReadDB:
         self,
         db_name,
         data,
-        update_type=config.DBUpdateType.SET,
+        update_type=DBUpdateType.SET,
         show_progress: bool = True,
         step: int = 10000,
         message="",
-        buff_limit=config.LMDB_BUFF_LIMIT,
+        buff_limit=LMDB_BUFF_LIMIT,
     ) -> bool:
         db = self.dbs[db_name]
         db_schema = dataclasses.asdict(self.db_schema[db_name])
@@ -533,7 +843,7 @@ class FReadDB:
                 p_bar.set_description(update_desc())
 
             db_obj = self.get_value(db_name, k)
-            if update_type == config.DBUpdateType.SET:
+            if update_type == DBUpdateType.SET:
                 if db_obj:
                     db_obj = set(db_obj)
                     v = set(v)
@@ -554,7 +864,7 @@ class FReadDB:
                 else:
                     c_new += 1
 
-            k, v = utils.serialize(
+            k, v = serialize(
                 k,
                 v,
                 **db_schema,
@@ -594,7 +904,7 @@ class FReadDB:
     def add(self, db_name, key, value) -> bool:
         db_schema = dataclasses.asdict(self.db_schema[db_name])
 
-        value = utils.serialize_value(value, **db_schema)
+        value = serialize_value(value, **db_schema)
         self.buff_size += len(value)
         self.buff[db_name].append([key, value])
         if self.buff_size > self.buff_limit:
